@@ -1,168 +1,9 @@
 ######## Multilevel Mokken scale analysis - MLweight and MLcoefH
 ######## Letty Koopman 
 ######## University of Amsterdam
-######## March 23, 2017
-
-# Used functions from the mokken package:
-"all.patterns" <- function(J,m){
-  grid <- list()
-  j <- 0;
-  p <- m^J
-  for (j in 1:J){
-    grid <- c(grid, j)
-    grid[[j]] <- 0:(m-1)
-  }
-  X <- t(expand.grid(grid))
-  dimnames(X) <- NULL
-  return(X[J:1,])
-}
-
-"phi" <- function(A,f, action){
-  # Numerical values are translations h(A %*% f) = A %*% f -
-  eps = 1E-80;
-  switch(action,
-         "identity" = A %*% f,
-         "exp"      = A %*% exp(f),
-         "log"      = A %*% log(abs(f)+eps),
-         "sqrt"     = A %*% sqrt(f),
-         "xlogx"    = A %*% (-f*log(f+eps)),
-         "xbarx"    = A %*% (f*(1-f))  # x(1-x)
-  )
-}
-
-"dphi" <- function(A,f,df, action){
-  eps=1E-80;
-  switch(action,
-         "identity" = A %*% df,
-         "exp"      = A %*% (as.numeric(exp(f)) * df),
-         "log"      = A %*% (as.numeric(1/(f+eps)) * df),
-         "sqrt"     = A %*% (as.numeric(1/(2*sqrt(f))) * df),
-         "xlogx"    = A %*% (as.numeric(-1-log(f+eps)) * df),
-         "xbarx"    = A %*% (as.numeric(1-2*f) * df)  #, x(1-x)
-  )
-}
-
-"check.data" <- function(X){
-  if (data.class(X) != "matrix" && data.class(X) != "data.frame")
-    stop("Data are not matrix or data.frame")
-  matrix.X <- as.matrix(X)
-  if (any(is.na(matrix.X))) stop("Missing values are not allowed")
-  if (mode(matrix.X)!="numeric") stop("Data must be numeric")
-  if (any(matrix.X < 0)) stop("All scores should be nonnegative")
-  if (any(matrix.X %% 1 !=0)) stop("All scores must be integers")
-  matrix.X <- matrix.X - min(matrix.X)
-  return(matrix.X)
-}
-
-
-## New functions to be implemented in mokken package
-"MLweight" <- function(X, maxx = NULL, minx = NULL){
-  # Computes the two-level Guttman weights for two-level Mokken Scale Analysis.
-  #
-  # Args:
-  #   X: Data matrix with a subject column and two item columns. The subject column is assumed to be the first.
-  #   maxx: The highest possible answer category. If not specified it is determined by using the lowest item score.
-  #   minx: The lowest possible answer category. If not specified it is determined by using the highest item score.
-  #   Depends on "all.patterns".
-  #
-  # Returns:
-  #   Guttman weights for two-level data.
-  
-  # Error handling:
-  if (ncol(X) != 3){
-    warning('X contains more than two items. Only first two items will be used')
-    X <- X[, 1:3]
-  }
-  if(is.null(minx)) minx <- min(X[, -1])
-  if(is.null(maxx)) maxx <- max(X[, -1])
-  X[, -1] <- X[, -1] - minx
-  maxx <- maxx - minx
-  minx <- 0
-  s <- X[, 1] # Subject column
-  f1 <- factor(X[, 2], levels = minx:maxx) # First item 
-  f2 <- factor(X[, 3], levels = minx:maxx) # Second item
-  
-  # Compute relative frequencies
-  Rel1 <- colSums(table(s, f1) / rowSums(table(s, f1))) 
-  names(Rel1) <- paste0(1, minx:maxx)
-  Rel2 <- colSums(table(s, f2) / rowSums(table(s, f2)))
-  names(Rel2) <- paste0(2, minx:maxx)
-  
-  # Cumulative relative frequencies, dealing with equal ranks
-  CumRel <- c(rev(cumsum(rev(Rel1[-1]))), rev(cumsum(rev(Rel2[-1]))))
-  names(CumRel) <- 1:length(CumRel)
-  y <- sort(CumRel, decreasing=T)
-  
-  if (any(duplicated(y))) {
-    perm <- function (n, r, v = 1:n, set = TRUE){
-      if (r == 1)  
-        matrix(v, n, 1)
-      else if (n == 1) 
-        matrix(v, 1, r)
-      else {
-        X <- NULL
-        for (i in 1:n) X <- rbind(X, cbind(v[i], Recall(n - 1, r - 1, v[-i])))
-        X
-      }
-    }
-    o <- lapply(unique(y), function(val) { 
-      m <- as.numeric(names(y[y==val]))
-      if (length(m) <= 1) {
-        return(m)
-      }
-      # do.call(paste0, as.data.frame(perm(length(m), length(m), m)))
-      as.data.frame(perm(length(m), length(m), m))
-    })
-    
-    # Ensure fixed ordering within items
-    for (i in 1:length(o)){
-      g <- o[[i]]
-      if(length(unique(g)) > 1){
-        select <- matrix(0, nrow(g))
-        for(j in 1:nrow(g)){
-          h <- g[j, ]
-          if(any(h >= 1 & h <= maxx) & any(h >= maxx + 1 & h <= maxx * 2)){
-            select[j] <- (all(h[which(h >= 1 & h <= maxx)] == sort(h[which(h >= 1 & h <= maxx)])) & all(h[which(h >= maxx + 1 & h <= maxx * 2)] == sort(h[which(h >= maxx + 1 & h <= maxx * 2)]))) * j
-          } else {
-            i1 <- ifelse(any(h >= 1 & h <= maxx), all(h[which(h >= 1 & h <= maxx)] == sort(h[which(h >= 1 & h <= maxx)])), 0)
-            i2 <- ifelse(any(h >= maxx + 1 & h <= maxx * 2), 
-                         all(h[which(h >= maxx + 1 & h <= maxx * 2)] == sort(h[which(h >= maxx + 1 & h <= maxx * 2)])), 0)
-            select[j] <- (i1 + i2) * j
-          }
-        }
-        o[[i]] <- matrix(apply(o[[i]][select, ], 1, paste0, collapse = ""))
-      }
-    }
-    
-    out <- matrix(do.call(paste0, expand.grid(o)))
-    w <- Z <- NULL
-    for(i in 1:nrow(out)){
-      ords <- as.numeric(unlist(strsplit(out[i, ],NULL)))
-      # Compute Z matrix for each item-response pattern
-      Z <- matrix(rep(matrix(all.patterns(2, maxx + 1), nrow = 1), maxx), nrow = maxx, byrow = TRUE)
-      Z <- matrix(ifelse(Z < row(Z), 0, 1), ncol = (maxx) * 2, byrow = TRUE)
-      Z <- Z[, ords]
-      
-      # Compute weights
-      w <- rbind(w, apply(Z, 1, function(x){sum(x * cumsum(abs(x - 1)))}))
-    }
-    wr <- matrix(colMeans(w), nrow = 1)
-  } else {
-    out <- paste(names(y), collapse = "")
-    
-    ords <- as.numeric(unlist(strsplit(out,NULL)))
-    # Compute Z matrix for each item-response pattern
-    Z <- matrix(rep(matrix(all.patterns(2, maxx + 1), nrow = 1), maxx), nrow = maxx, byrow = TRUE)
-    Z <- matrix(ifelse(Z < row(Z), 0, 1), ncol = (maxx) * 2, byrow = TRUE)
-    Z <- Z[, ords]
-    
-    # Compute weights
-    wr <- apply(Z, 1, function(x){sum(x * cumsum(abs(x - 1)))})
-    
-  }
-  return(wr)
-}
+######## June 30, 2017
 
+# Overdispersion covariance matrix including the shortcut
 "MLcoefH" <- function(X, se = TRUE, nice.output = TRUE, subject = 1){
   # Computes the two-level scalability coefficients in Mokken scale analysis
   #
@@ -180,6 +21,7 @@
   #   Two-level scalability coefficients and optionally their standard errors.
   
   # Error handling:
+  eps <- 1E-80
   if(subject != 1){
     X <- cbind(X[, subject], X[, -subject])
   }
@@ -207,8 +49,10 @@
   J <- ncol(X[, -1])
   K <- choose(J, 2) 
   g <- m + 1
+  B <- K * g^2 
+  U <- J * g 
   
-  nams <- apply(combn(colnames(X)[-1],2), 2, function(z) paste(z, collapse = ' ')) 
+  nams <- apply(combn(colnames(X)[-1],2), 2, paste, collapse = ' ') 
   cols <- combn(J, 2)
   Patterns <- cbind("Xa" = rep(0:m, each = m + 1), "Xb" = rep(0:m, m + 1)) 
   
@@ -216,179 +60,242 @@
   
   if(se == TRUE){
     R <- unique(X)
-    n <- as.numeric(table(factor(apply(X, 1, paste, collapse=","), levels=unique(apply(X, 1, paste, collapse=",")))))
-    Rss <- table(R[, 1])
+    Rss <- RRs <- table(R[, 1])
     Rd <- rep(Rs, Rss)
+    n <- as.numeric(table(factor(apply(X, 1, paste, collapse=","), levels=unique(apply(X, 1, paste, collapse=","))))) 
+    Subs <- R[, 1]
+    SubsX <- X[, 1]
+    Rt <- aggregate(n ~ ., data = R[, -1], sum) 
+    np <- Rt[, J + 1] 
+    Lst <- length(np)
+    Rt <- Rt[, -(J + 1)]
+    nprel <- aggregate(n / Rd ~ ., data = R[, -1], sum)[, J + 1]
+    npbet <- aggregate(n / (Rd * (Rd - 1)) ~ ., data = R[, -1], sum)[, J + 1]
+    nN <- cbind(R, rownr = 1:nrow(R))
+    nN <- aggregate(rownr ~ ., nN[, -1], unique)[, J + 1]
     
-    G5 <- matrix(0, 3 * K, nrow(R))
-    g5 <- matrix(0, K, 3)
-    ABs <- AWs <- AEs <- matrix(0, g^2, nrow(R))
-    Fw <- Fb <- Fe <- NULL
-      
+    # Covariance matrix with correction for overdispersion
+    patsX <- apply(X[, -1], 1, paste, collapse = "")
+    patsR <- apply(Rt, 1, paste, collapse = "")
+    pSubs <- matrix(0, LS, Lst)
+    for(s in S) {
+      pSubs[s, ] <- rowSums(outer(patsR, patsX[SubsX == s], "==")) / Rs[s]
+    }
+    covps <- cov(pSubs)
+    p <- colMeans(pSubs)
+    covp <- (diag(length(p)) * p - (p %*% t(p))) # of hier dan ook S * nu?
+    nu <- 1/(sum(1/Rs) / LS)
+    
+    covtot <- LS * nu * (nu - 1) * covps + LS * nu * covp
+    
+    # shortcut, deel kopie van hieronder
+    Umat <- matrix(0, U, length(np))
+    row <- 0
+    ns <- list()
+    for(i in 1:J) {
+      Subs <- R[, 1]
+      Xa <- R[, i + 1]
+      Ta <- Rt[, i]
+      ns[[i]] <- matrix(0, m + 1, LS)
+      for(a in 0:m){
+        row <- row + 1
+        ns[[i]][a + 1, ] <- tapply((Xa == a) * n, Subs, sum)
+        nit <- (Ta == a) * nprel
+        Umat[row, ] <- nit / np
+      }
+    }
+    
+    Smat <- (aggregate(n / Rd ~ ., data = R[, -1], sum)[, J + 1]) / np
+    # bij gelijke groepsgrootte lijkt dit 1 / Rs
+    
+    Fwt <- Fbt <- Fet <- Fw <- Fb <- Fe <- eij <- NULL
+    G3W <- G3B <- G3E <- matrix(0, K, length(np))
     for(k in 1:K){
       z <- cols[, k]
-      Subs <- R[, 1]
-      Xa <- R[, z[1] + 1]
-      Xb <- R[, z[2] + 1]
-      Weights <- MLweight(X[, c(1, z + 1)], minx = 0, maxx = m)# Weights <- MLweight(Rsub, minx = 0, maxx = m)
+      Subs <- X[, 1]
+      Xa <- X[, z[1] + 1]
+      Xb <- X[, z[2] + 1]
+      Ra <- R[, z[1] + 1]
+      Rb <- R[, z[2] + 1]
+      Ta <- Rt[, z[1]]
+      Tb <- Rt[, z[2]]
+      Rss <- rep(Rs, Rs)
+      Weights <- MLweight(X[, c(1, z + 1)], minx = 0, maxx = m)
+      Wmat <- Bmat <- Emat <- matrix(0, g^2, length(np))
+      
       for(x in 1:g^2){
         if(Weights[x] > 0){
           i <- Patterns[x, 1]
           j <- Patterns[x, 2]
-          nw <- tapply((Xa == i & Xb == j) * n, Subs, sum)
-          ni <- tapply((Xa == i) * n, Subs, sum)
-          Ni <- sum(ni / Rs)
-          nj <- tapply((Xb == j) * n, Subs, sum)
-          Nj <- sum(nj / Rs)
-          at <- (Xa == i) * (rep(nj, Rss) - (Xb == j))
-          nb <- tapply(at * n, Subs, sum)
           
-          Fw[x] <- Weights[x] * sum(nw / Rs)
-          Fb[x] <- Weights[x] * sum(nb / (Rs * (Rs - 1)))
-          Fe[x] <- Weights[x] * Ni * Nj / LS
+          #at <- (Xa == i) * (rep(nj, Rs) - (Xb == j))
+          nJJJ <- nIII <- nj <- nb <- NULL
+          for(s in S) {
+            nJJJ[s] <- sum(Xa[Subs == s] == i & Xb[Subs == s] == j)
+            nitemp <- Xa[Subs == s] == i
+            nIII[s] <- sum(nitemp)
+            njtemp <- Xb[Subs == s] == j
+            nj[s] <- sum(njtemp)
+            nb[s] <- sum(nitemp * (nj[s] - njtemp))
+          }
+          bijJ <- nJJJ > 0
+          nw <- sum(nJJJ / Rs)
+          ni <- sum(nIII / Rs)
+          bijI <- nIII > 0
+          at <- (Ra == i) * (rep(nj, RRs) - (Rb == j)) / (rep(Rs, RRs) * (rep(Rs, RRs) - 1)) * n
+          at <- sapply(1:length(np), function(x) sum(at[nN[[x]]]))
+          nj <- sum(nj / Rs)
           
-          cat <- as.numeric(Xa == i) + as.numeric(Xb == j)
-          ABs[x, ] <- Weights[x] * (at * Rd - rep(nb, Rss)) / (Rd^2 * (Rd - 1))
-          AWs[x, ] <- Weights[x] * ifelse(cat == 2, (Rd - rep(nw, Rss)) / Rd^2, -rep(nw, Rss) / Rd^2)
-          AEs[x, ] <- Weights[x] * ifelse(cat == 2, (Rd * (Ni + Nj) - Ni * rep(nj, Rss) - Nj * rep(ni, Rss)) / (Rd^2 * LS), 
-                                           ifelse(cat == 1 & Xa == i, -(Ni * rep(nj, Rss) + Nj * (rep(ni, Rss) - Rd)) / (Rd^2 * LS), 
-                                                  ifelse(cat == 1 & Xb == j, -(Nj * rep(ni, Rss) + Ni * (rep(nj, Rss) - Rd)) / (Rd^2 * LS),
-                                                         (-Ni * rep(nj, Rss) - Nj * rep(ni, Rss)) / (Rd^2 * LS))))
-         
+          wt <- (Ta == i & Tb == j) * nprel #/ Rd
+          eij <- ni * nj / LS
+          
+          nit <- (Ta == i) * nprel
+          njt <- (Tb == j) * nprel
+          
+          Wmat[x, ] <- (wt / np) * Weights[x]
+          Bmat[x, ] <- (at / np) * Weights[x]
+          Emat[x, ] <- Weights[x] * ((Ta == i) * (nit / np) * eij / (ni + eps) + (Tb == j) * (njt/np) * eij / (nj + eps) - eij * Smat / LS ) 
+          Fwt[x] <- Weights[x] * nw
+          Fbt[x] <- Weights[x] * sum(nb / (Rs * (Rs - 1)))
+          Fet[x] <- Weights[x] * eij
         } else {
-          Fw[x] <- Fb[x] <- Fe[x] <- 0
-          ABs[x, ] <- AWs[x, ] <- AEs[x, ] <- 0
+          Fwt[x] <- Fbt[x] <- Fet[x] <- 0
         }
-      }
-      g5[k, 1] <- sum(Fb)
-      g5[k, 2] <- sum(Fw)
-      g5[k, 3] <- sum(Fe)
+      } 
       
-      G5[k, ] <- colSums(ABs)
-      G5[k + K, ] <- colSums(AWs)
-      G5[k + 2 * K, ] <- colSums(AEs)
+      G3W[k, ] <- colSums(Wmat)
+      G3B[k, ] <- colSums(Bmat)
+      G3E[k, ] <- colSums(Emat)
+      Fw[k] <- sum(Fwt)
+      Fb[k] <- sum(Fbt)
+      Fe[k] <- sum(Fet)
     }
-
-    g5i <- matrix(0, J, 3)
+    
+    # Hij, Hi and H coefficients
+    Fwi <- Fbi <- Fei <- NULL
+    G3Wi <- G3Bi <- G3Ei <- matrix(0, J, length(np))
     for(i in 1:J) {
-      g5i[i, ] <- colSums(g5[apply(cols, 2, function(x) any(x == i)), ])
+      items <- apply(cols, 2, function(x) any(x == i))
+      Fwi[i] <- sum(Fw[items])
+      Fbi[i] <- sum(Fb[items])
+      Fei[i] <- sum(Fe[items])
+      G3Wi[i, ] <- colSums(G3W[items, ])
+      G3Bi[i, ] <- colSums(G3B[items, ])
+      G3Ei[i, ] <- colSums(G3E[items, ])
     }
-    g5i <- rbind(g5i[1, 1], matrix(g5i))
     
-    g5ii <- colSums(g5)
-    g5ii <- rbind(g5ii[1], matrix(g5ii))
+    g3 <- matrix(c(Fb[1], Fb, Fw, Fe))
     
-    g5 <- rbind(g5[1, 1], matrix(g5))
-   
-    # for item coefficients
-    G5i <- matrix(0, J * 3, nrow(R))
-    for(i in 1:J) {
-      items <- which(apply(cols, 2, function(x) any(x == i)))
-      G5i[i, ] <- colSums(G5[items, ])
-      G5i[i + J, ] <- colSums(G5[items + K, ])
-      G5i[i + J * 2, ] <- colSums(G5[items + K * 2, ])
-    }
-    G5i <- rbind(G5i[1, ], G5i)
+    g3i <- matrix(c(Fbi[1], Fbi, Fwi, Fei))
     
-    # for total scale coefficients
-    G5ii <- rbind(colSums(G5[1:K, ]), colSums(G5[1:K, ]), 
-                  colSums(G5[(K + 1):(2 * K), ]), colSums(G5[(2 * K + 1):(3 * K), ]))
-
-    G5 <- rbind(G5[1, ], G5)
+    g3ii <- matrix(c(sum(Fb), sum(Fb), sum(Fw), sum(Fe)))
+    
+    # Jacobian Hij, Hi and H coefficients
+    G3 <- rbind(G3B[1, ], G3B, G3W, G3E)
+    
+    G3i <- rbind(G3Bi[1, ], G3Bi, G3Wi, G3Ei)
+    
+    G3ii <- rbind(colSums(G3B), colSums(G3B), colSums(G3W), colSums(G3E))
+    
+    ## Remaining design matrices
     
     # Create A6 --> To compute the ratio of observed to expected errors
-    A6 <- rbind(c(1, -1, rep(0, 3 * K - 1)), 
-                cbind(rep(0, 2 * K), diag(2 * K), 
-                      rbind(-diag(K), -diag(K))))
-    
-    A6i <- rbind(c(1, -1, rep(0, 3 * J - 1)), 
-                 cbind(rep(0, 2 * J), diag(2 * J), 
-                       rbind(-diag(J), -diag(J))))
-    
-    A6ii <- matrix(c(1, 0, 0, -1, 1, 0, 0, 0, 1, 0, -1, -1), 3)
+    A6 <- rbind(matrix(c(1, -1, rep(0, 3 * K - 1)), 1), 
+                cbind(matrix(0, K * 2, 1), diag(K * 2), rbind(-1 * diag(K), -1 * diag(K))))
+    A6i <- rbind(matrix(c(1, -1, rep(0, 3 * J - 1)), 1), 
+                 cbind(matrix(0, J * 2, 1), diag(J * 2), rbind(-1 * diag(J), -1 * diag(J))))
+    A6ii <- rbind(matrix(c(1, -1, rep(0, 3 - 1)), 1), 
+                  cbind(matrix(0, 2, 1), diag(2), rbind(-1, -1)))
     
     # Create A7 --> To compute the Hij/Hi/H values
-    A7 <- cbind(rep(1, 2 * K), -diag(2 * K))
-    
-    A7i <- cbind(rep(1, 2 * J), -diag(2 * J))
-    
-    A7ii <- matrix(c(1, 1, -1, 0, 0, -1), 2)
+    A7 <- rbind(cbind(rep(1, 2 * K), -diag(2 * K)),
+                cbind(rep(1, 2 * K), -diag(2 * K)))
+    A7i <- rbind(cbind(rep(1, 2 * J), -diag(2 * J)),
+                 cbind(rep(1, 2 * J), -diag(2 * J)))
+    A7ii <- matrix(c(1, 1, 1, 1, -1, 0, -1, 0, 0, -1, 0, -1), 4)
     
     # Create A8 and A9 --> To compute ratio HB/HW
-    A8 <- cbind(diag(K), -diag(K))
-    A8i <- cbind(diag(J), -diag(J))
-    A8ii <- matrix(c(1, -1), 1)
+    A8 <- cbind(diag(K * 3), rbind(matrix(0, 2 * K, K), -diag(K)))
+    A8i <- cbind(diag(J * 3), rbind(matrix(0, 2 * J, J), -diag(J)))
+    A8ii <- matrix(c(1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, -1), 3)
     
-    A9 <- diag(K)
-    A9i <- diag(J)
-    A9ii <- matrix(1)
+    A9 <- diag(K * 3)
+    A9i <- diag(J * 3)
+    A9ii <- diag(3)
     
-    #Hij
+    # Hij
     
-    g6 <- phi(A6, g5, "log")
-    G6 <- dphi(A6, g5, G5, "log")
+    g4 <- phi(A6, g3, "log")
+    G4 <- dphi(A6, g3, G3, "log")
     
-    g7 <- phi(A7, g6, "exp") 
-    G7 <- dphi(A7, g6, G6, "exp")
+    g5 <- phi(A7, g4, "exp")
+    G5 <- dphi(A7, g4, G4, "exp")
     
-    g8 <- phi(A8, g7, "log") 
-    G8 <- dphi(A8, g7, G7, "log")
+    g6 <- phi(A8, g5, "log")
+    G6 <- dphi(A8, g5, G5, "log")
     
-    G9 <- dphi(A9, g8, G8, "exp")
+    g7 <- phi(A9, g6, "exp")
+    G7 <- dphi(A9, g6, G6, "exp")
     
-    HBij <- g7[1:K, ]
-    HWij <- g7[-c(1:K), ]
+    se.Hij <- sqrt(diag(G7 %*% (covtot %*% t(G7))))# sqrt(diag(G7 %*% (as.numeric(np) * t(G7))))
+    
+    HBij <- g5[1:K, ]
+    HWij <- g5[(K + 1):(K * 2), ]
     HBWij <- HBij / HWij
     
-    se.Hij <- sqrt(diag(G7 %*% (as.numeric(n) * t(G7))))
     se.HBij <- se.Hij[1:K]
-    se.HWij <- se.Hij[-c(1:K)]
-    se.HBWij <- sqrt(diag(G9 %*% (as.numeric(n) * t(G9))))
+    se.HWij <- se.Hij[(K + 1):(K * 2)]
+    se.HBWij <- se.Hij[-c(1:(K * 2))]
+    
     
     # Hi
     
-    g6 <- phi(A6i, g5i, "log") 
-    G6 <- dphi(A6i, g5i, G5i, "log")
+    g4 <- phi(A6i, g3i, "log")
+    G4 <- dphi(A6i, g3i, G3i, "log")
     
-    g7 <- phi(A7i, g6, "exp")
-    G7 <- dphi(A7i, g6, G6, "exp")
+    g5 <- phi(A7i, g4, "exp")
+    G5 <- dphi(A7i, g4, G4, "exp")
     
-    g8 <- phi(A8i, g7, "log") 
-    G8 <- dphi(A8i, g7, G7, "log")
+    g6 <- phi(A8i, g5, "log")
+    G6 <- dphi(A8i, g5, G5, "log")
     
-    g9 <- phi(A9i, g8, "exp") 
-    G9 <- dphi(A9i, g8, G8, "exp")
+    g7 <- phi(A9i, g6, "exp")
+    G7 <- dphi(A9i, g6, G6, "exp")
     
-    HBi <- g7[1:J, ]
-    HWi <- g7[-c(1:J), ]
+    se.Hi <- sqrt(diag(G7 %*% (covtot %*% t(G7))))
+    
+    HBi <- g5[1:J, ]
+    HWi <- g5[(J + 1):(J * 2), ]
     HBWi <- HBi / HWi
     
-    se.Hi <- sqrt(diag(G7 %*% (as.numeric(n) * t(G7))))
     se.HBi <- se.Hi[1:J]
-    se.HWi <- se.Hi[-c(1:J)]
-    se.HBWi <- sqrt(diag(G9 %*% (as.numeric(n) * t(G9))))
+    se.HWi <- se.Hi[(J + 1):(J * 2)]
+    se.HBWi <- se.Hi[-c(1:(J * 2))]
     
     # H
     
-    g6 <- phi(A6ii, g5ii, "log")
-    G6 <- dphi(A6ii, g5ii, G5ii, "log")
+    g4 <- phi(A6ii, g3ii, "log")
+    G4 <- dphi(A6ii, g3ii, G3ii, "log")
     
-    g7 <- phi(A7ii, g6, "exp")
-    G7 <- dphi(A7ii, g6, G6, "exp")
+    g5 <- phi(A7ii, g4, "exp")
+    G5 <- dphi(A7ii, g4, G4, "exp")
     
-    g8 <- phi(A8ii, g7, "log") 
-    G8 <- dphi(A8ii, g7, G7, "log")
+    g6 <- phi(A8ii, g5, "log")
+    G6 <- dphi(A8ii, g5, G5, "log")
     
-    g9 <- phi(A9ii, g8, "exp") 
-    G9 <- dphi(A9ii, g8, G8, "exp")
+    g7 <- phi(A9ii, g6, "exp")
+    G7 <- dphi(A9ii, g6, G6, "exp")
     
-    HB <- g7[1, ]
-    HW <- g7[2, ]
+    se.H <- sqrt(diag(G7 %*% (covtot %*% t(G7))))
+    
+    HB <- g5[1, ]
+    HW <- g5[2, ]
     HBW <- HB / HW
-    se.H <- sqrt(diag(G7 %*% (as.numeric(n) * t(G7))))
+    
     se.HB <- se.H[1]
     se.HW <- se.H[2]
-    se.HBW <- sqrt(diag(G9 %*% (as.numeric(n) * t(G9))))
+    se.HBW <- se.H[3]
+    
     
     if(nice.output == TRUE){
       Hij <- HBijt <- matrix(0, J, J, dimnames = list(labels, labels))
@@ -553,7 +460,8 @@
     } 
   }
   
-
+  
   return(OL)
   
-  }
+}
+
